#!/usr/bin/env python3
from ev3dev2.motor       import (MoveTank, LargeMotor, MediumMotor,
                                 OUTPUT_A, OUTPUT_B, OUTPUT_C, OUTPUT_D)
import socket
import threading

HOST    = ''
PORT    = 2334
SIZE    = 32

class Robot:

    def __init__(self):
        self.ruedas          = MoveTank(OUTPUT_C, OUTPUT_A)
        self.viraje          = MediumMotor(OUTPUT_D)
        self.brazo           = LargeMotor(OUTPUT_B)

        self.service         = threading.Thread(target=self.daemon, args=(), daemon=True).start()

    def avanzar(self, velocidad=10):
        self.ruedas.on(left_speed=-velocidad, right_speed=-velocidad)

    def retroceder(self, velocidad=10):
        self.ruedas.on(left_speed=velocidad, right_speed=velocidad)
    
    def izquierda(self, velocidad=10):
        self.viraje.on(speed=velocidad)

    def derecha(self, velocidad=10):
        self.viraje.on(speed=-velocidad)

    def atacar(self, velocidad=10, rotation=True):
        if rotation:
            self.brazo.on_for_rotations(speed=velocidad, rotations=1)
        else:
            self.brazo.on(speed=velocidad)

    def perdonar(self):
        self.brazo.off()

    def apagar(self):
        self.ruedas.off()
        self.brazo.off()
        self.viraje.off()

    def detener(self):
        self.ruedas.off()

    def daemon(self):
        while (True):
            if(self.viraje.is_stalled): #Destraba las ruedas
                self.viraje.reset()

if __name__ == "__main__":
    print("Inicializando servidor...")
    robot = Robot()
    exit  = False
    while True:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            s.bind((HOST, PORT))
            s.listen(5)
            while True:
                try:
                    print("Esperando Cliente...")
                    clientsocket, address = s.accept()
                    with clientsocket:
                        print("Conectado a {}".format(address))
                        data = None
                        key  = None
                        while True:
                            data     = clientsocket.recv(SIZE).decode("utf-8").split(" ")
                            key      = data[0]
                            if data[1] != "":
                                speed = float(data[1])
                                if data[2] != "":
                                    rotation = int(data[2])
                                
                            if key == "apagar":
                                robot.apagar()
                            elif key == "atacar":
                                robot.atacar(speed, rotation)
                            elif key == "avanzar":
                                robot.avanzar(speed)
                            elif key == "derecha":
                                robot.derecha(speed)
                            elif key == "detener":
                                robot.detener()
                            elif key == "izquierda":
                                robot.izquierda(speed)
                            elif key == "perdonar":
                                robot.perdonar()
                            elif key == "retroceder":
                                robot.retroceder(speed)
                            elif key == "pass": # Para comprobar la conexión
                                pass
                            else:
                                print("Clave desconocida: {}".format(data[0]))
                                print("Desconectando...")
                                break; #Acepta Nuevo Cliente
                except socket.error as e:
                    print("Error: {}".format(str(e)))
                    print("Cliente desconectado")
                except IndexError as e:
                    print("Error: {}".format(str(e)))
                    print("Cliente desconectado")
                except KeyboardInterrupt:
                    exit = True
                    break;
                except Exception as e:
                    print("Error")
                    print(str(e))
                    exit = False
                    break; #Reiniciar
                robot.apagar()
        robot.apagar()
        if exit:
            print("\nSaliendo...")
            break; #Salir
        else:
            print("Reiniciando...")