from numpy import sin, cos, sqrt, deg2rad, pi

# Asumiendo que es un choque plástico
# Vt = Vi
# ^ El choque producido por el brazo no es plástico.

# Funciones Físicas
cVy    = lambda v, θ : v*sin(θ)
cVx    = lambda v, θ : v*cos(θ)

def cTv(vi, θ) -> float:
    """
        Calcular Tiempo de Vuelo.\n
        Retorna el tiempo de vuelo en segundos

        Parámetros:
            θ:  Ángulo en Grados
            vi: Velocidad Inicial en m/s
        
        Retorna:
            Tiempo de Vuelo
    """
    β  = deg2rad(θ)
    Vy = cVy(vi, β)
    # Raíz: Rama positiva = 2° punto de intersección con el Yf en la parábola asociada (caída).
    Tv = (Vy+sqrt(Vy**2-2*G*ΔY))/G
    return Tv

def cVbP(d, θ) -> float:
    """
        Calcular Velocidad Brazo en Porcentaje\n
        Retorna la velocidad del brazo necesaria para recorrer la distancia solicitada

        Parámetros:
            θ: Ángulo en Grados
            d: Distancia en metros
        
        Retorna:
            Velocidad Porcentual del Brazo
    """
    Vb = Vt2Per(cVi(d, θ))
    return Vb

def cVi(d, θ) -> float:
    """
        Calcular Velocidad Inicial\n
        Retorna la velocidad inicial necesaria para que la pelota llegue al objetivo
        
        Parámetros:
            θ: Ángulo en Grados
            d: Distancia en Metros

        Retorna:
            Velocidad Inicial
    """
    ΔX = d
    β  = deg2rad(θ)
    vi = sqrt(G*(ΔX)**2/(2*cos(β)*(ΔX*sin(β)-ΔY*cos(β))))
    return vi

def cD(vi, θ) -> float:
    """
        Calcular Distancia\n
        Retorna la distancia entre la posición final e inicial en el eje X en metros

        Parámetros:
            θ:  Ángulo en Grados.
            vi: Velocidad Inicial.

        Retorna:
            Distancia
    """
    β  = deg2rad(θ)
    Vx = cVx(vi, β)

    t = cTv(vi, θ)
    d  = Vx*t
    return d

# Convertidores
# # Definiciones:
# # • Per -> Velocidad Porcentual del Brazo
# # • RPM -> Revoluciones por Minuto del Brazo
# # • Vt  -> Velocidad Tangencial del Brazo

def Per2RPM(per):
    if not -100 <= per <= +100:
        raise Exception("Valor debe ser entre [-100, +100]")
    return per*maxRPM/100
def Per2Vt(per):
    return RPM2Vt(Per2RPM(per))

def RPM2Per(rpm):
    if not -maxRPM <= rpm <= maxRPM:
        raise Exception(f"Valor debe ser entre [{-maxRPM} y {maxRPM}]")
    return rpm*100/maxRPM
def RPM2Vt(rpm):
    if not -maxRPM <= rpm <= maxRPM:
        raise Exception(f"Valor debe ser entre [{-maxRPM} y {maxRPM}]")
    return rpm*pi*R/30

def Vt2Per(Vt):
    return RPM2Per(Vt2RPM(Vt))
def Vt2RPM(Vt):
    if not -maxVt <= Vt <= maxVt:
        raise Exception(f"Valor debe ser entre [{-maxVt} y {maxVt}]")
    return 30*Vt/(R*pi)

# Datos Iniciales
G  = 9.81                   # Aceleración de la Gravedad (Metros/Segundo^2)
Yi = 0.05                   # Altura de la Plataforma    (Metros)
Yf = 0                      # Altura del Suelo           (Metros)
ΔY = Yf-Yi

# Robot Brazo (LargeMotor)
R        = 0.15             # Radio del Brazo (Metros)
maxRPM   = 175              # RPM máximo del brazo, según reportes de ev3dev2
maxVt    = RPM2Vt(maxRPM)
maxD     = cD(maxVt, 45)

if __name__ == "__main__":
    # Pruebas
    θ  = 45                 # Ángulo    (Grados)
    d  = 1                  # Distancia (Metros)
    
    v = cVi(d, θ)
    print(f"Ángulo:\t\t\t{θ}°\nDistancia:\t\t{d}m")
    print("\nProbando cVi y cD:")
    print(f"Velocidad Máx:\t\t{maxVt}\tm/s")
    print(f"Velocidad:\t\t{v}\tm/s\nDistancia:\t\t{cD(v, θ)}\tmetros")

    #8.5
    print(f"\nNo Relacionado:\n Distancia Máxima: {maxD}")