#!/usr/bin/env python3
from   Física.Convertidor    import dis2vel
from   Conexión.Robot_Client import Robot
import threading
import tkinter               as     tk
from   tkinter               import ttk
from   PIL                   import Image, ImageTk
from   functools             import wraps
from   platform              import system

Linux = True if system() == "Linux" else False
if Linux:
    from Manilla.Gamepad_Process_Pipe import Gamepad, get_gamepads, check
    from queue                        import Queue
    from ttkthemes                    import ThemedTk

TITULO = "Tigerbot"

def Tk():
    ventana = None
    if Linux:
        # adapta
        ventana = ThemedTk(className=TITULO, theme="plastik")

    else:
        ventana = tk.Tk(className=TITULO)
    return ventana

def hilo(func):
    @wraps(func)
    def run(*args, **kwargs):
        t = threading.Thread(target=func, args=args, kwargs=kwargs, daemon=True)
        t.start()
        return t
    return run

def dict_velocidad(parent: tk.Widget) -> (dict[str, dict[str, ttk.Widget]], tk.IntVar, tk.IntVar):
    mov_sli_val = tk.DoubleVar(value=50)
    gol_sli_val = tk.DoubleVar(value=50)
    mov_lab_val = tk.IntVar(value=int(mov_sli_val.get()))
    gol_lab_val = tk.IntVar(value=int(gol_sli_val.get()))
    
    """Cambia el valor del label (entero) dependiendo del valor del slider (double)"""
    mov_sli_val.trace_add("write", lambda x,y,z:mov_lab_val.set(int(mov_sli_val.get())))
    gol_sli_val.trace_add("write", lambda x,y,z:gol_lab_val.set(int(gol_sli_val.get())))
    
    return {
        'movimiento': {
            'label'       : ttk.Label(parent, text="Velocidad"),
            'label_value' : ttk.Label(parent, textvariable=mov_lab_val),
            'slider'      : ttk.Scale(parent, from_=0, to=100, variable=mov_sli_val, orient=tk.HORIZONTAL)
            #,'value'       : mov_lab_val
            },
        'golpe'     : {
            'label'       : ttk.Label(parent, text="Golpe"),
            'label_value' : ttk.Label(parent, textvariable=gol_lab_val),
            'slider'      : ttk.Scale(parent, from_=0, to=100, variable=gol_sli_val, orient=tk.HORIZONTAL)
            #,'rvalue'    : gol_lab_val
            #, 'wvalue'   : gol_sli_val
            }
        }, mov_lab_val, gol_lab_val, gol_sli_val
    # Si quisiera modificar el valor de otras formas, tendría que reemplazar el x_lab_val con x_sli_val
def dict_control(parent: tk.Widget) -> dict[str, ttk.Button]:
    return {
            'up'    : ttk.Button(parent, text="↑"),
            'left'  : ttk.Button(parent, text="←"),
            'down'  : ttk.Button(parent, text="↓"),
            'right' : ttk.Button(parent, text="→"),
            'center': ttk.Button(parent, text="0")
        }
def dict_mando(parent: tk.Widget) -> dict[str, ttk.Widget]:
    return {
            'label' : ttk.Label(parent, text="Mando:"),
            'combo' : ttk.Combobox(parent, values=None, state="readonly"),
            'button': ttk.Button(parent, text="Seleccionar")
    }
def dict_mecánica(parent: tk.Widget) -> dict[str, ttk.Widget]:
    return {
            'label' : ttk.Label(parent, text="Centímetros:"),
            'entry' : ttk.Entry(parent),
            'button': ttk.Button(parent, text="☑")
    }
def dict_ip(parent: tk.Widget) -> dict[str, ttk.Widget]:
    return {
            'label' : ttk.Label(parent, text="IP"),
            'entry' : ttk.Entry(parent),
            'button': ttk.Button(parent, text="☑")
    }

class Interfaz():
    def __init__(self, ventana_principal: tk.Tk):
        #Propiedades de la Ventana
        ventana_principal.geometry("600x400")
        ventana_principal.title(TITULO)
        ventana_principal.resizable(False, False)
        
        #Variables del objeto
        self.logo:      tk.PhotoImage = ImageTk.PhotoImage(Image.open("Imágenes/tigerbot1.png").resize((100, 100)))
        self.ico:       tk.PhotoImage = ImageTk.PhotoImage(Image.open("Imágenes/tigerbot0.png").resize((100, 100)))
        self.ventana:   tk.Tk         = ventana_principal
        self.ventana.wm_iconphoto(True, self.ico)
        
        #Escena
        self.root: ttk.Frame = None

        #Control Robot
        self.robot:    Robot = Robot()
        
        #Diccionarios
        self.velocidad: dict = None
        self.control:   dict = None
        self.mando:     dict = None
        self.mecánica:  dict = None
        self.ip:        dict = None
        
        #Mando
        if Linux:
            self.manilla_actual: Gamepad = None
            self.comunicación  : Queue   = Queue(maxsize=5)

        self.crear_gui()
        self.asignar_eventos()
        if Linux:
            self.monitor1()

    def crear_marco(self): # Innecesario
        """Crea o recrea un marco que ocupa toda la ventana"""
        if self.root:
            self.root.destroy()
        self.root = ttk.Frame(self.ventana)
        self.root.pack(fill=tk.BOTH, expand=True)

    def crear_gui(self):
        def crear_top(parent: ttk.Frame) -> ttk.Frame:
            top_frame = ttk.Frame(parent)
            top_frame.rowconfigure(0, weight=1)
            
            top_frame.columnconfigure(0, weight=1)
            velocidad_frame = ttk.Frame(top_frame)
            velocidad_frame.grid(row=0, column=0, sticky=tk.NSEW)

            velocidad, vel_mov, r_vel_gol, w_vel_gol = dict_velocidad(velocidad_frame)
            
            # Posiciona y les da espacio a cada widget en el diccionario
            for row, vel_dict in enumerate(velocidad.values()):
                velocidad_frame.rowconfigure(row, weight=1)
                for column, widget in enumerate(vel_dict.values()):
                    widget: ttk.Widget
                    if row == 0:
                        if column == 0:
                            velocidad_frame.columnconfigure(column, weight=1)
                            widget.grid(row=row, column=column, sticky=tk.SW, padx=(5, 0))
                        else:
                            widget.grid(row=row, column=column, sticky=tk.SE)
                    else:
                        if column == 0:
                            velocidad_frame.columnconfigure(column, weight=1)
                            widget.grid(row=row, column=column, sticky=tk.NW, padx=(5,0))
                        else:
                            widget.grid(row=row, column=column, sticky=tk.NE)


            #Agrega las claves después de poner los widgets
            velocidad['movimiento'].update({'value': vel_mov})
            velocidad['golpe'].update({'r_value': r_vel_gol})
            velocidad['golpe'].update({'w_value': w_vel_gol})
            
            ttk.Label(top_frame, image=self.logo).grid(row=0, column=1, sticky=tk.NE)
            
            self.velocidad = velocidad
            
            return top_frame
        def crear_middle(parent):
            middle_frame = ttk.Frame(parent)
            middle_frame.rowconfigure(0, weight=1)
            
            middle_frame.columnconfigure((0, 1), weight=1)
            left_frame = ttk.Frame(middle_frame)
            left_frame.grid(row=0, column=0, sticky=tk.NSEW)
            right_frame = ttk.Frame(middle_frame)
            right_frame.grid(row=0, column=1, sticky=tk.NSEW)
            
            #Middle-Left
            control = dict_control(left_frame)
            left_frame.rowconfigure((0, 1, 2), weight=1)
            left_frame.columnconfigure((0, 1, 2), weight=1)
            
            control['up'].grid(row=0, column=1, sticky=tk.NSEW, padx=5, pady=(5, 2.5))
            control['left'].grid(row=1, column=0, sticky=tk.NSEW, padx=(5, 2.5), pady=5)
            control['down'].grid(row=2, column=1, sticky=tk.NSEW, padx=5, pady=(2.5, 5))
            control['right'].grid(row=1, column=2, sticky=tk.NSEW, padx=(2.5, 5), pady=5)
            control['center'].grid(row=1, column=1, sticky=tk.NSEW, padx=2.5, pady=2.5, ipadx=2.5, ipady=2.5)
            
            #Middle-Right
            right_frame.columnconfigure(1, weight=1)
            right_frame.config(padding=(5,5,5,5))

            mecánica = dict_mecánica(right_frame)
            mecánica['label'].grid(row=0, column=0, sticky=tk.NSEW, padx=(0,2))
            mecánica['entry'].grid(row=0, column=1, sticky=tk.NSEW, padx=(2,2))
            mecánica['button'].grid(row=0, column=2, sticky=tk.NSEW, padx=(2,0))
            
            #Columna intermedia de tamaño 5
            right_frame.rowconfigure(1, minsize=5)
            #Columna que ocupa todo el espacio libre
            right_frame.rowconfigure(3, weight=1)
            
            if Linux:
                mando = dict_mando(right_frame)
                mando['label'].grid(row=2, column=0, sticky=tk.NSEW, padx=(0,2))
                mando['combo'].grid(row=2, column=1, sticky=tk.NSEW, padx=(2,2))
                mando['button'].grid(row=2, column=2, sticky=tk.NSEW, padx=(2,0))
                self.mando = mando

            self.control  = control
            self.mecánica = mecánica
            return middle_frame
        def crear_bottom(parent):
            bottom_frame = ttk.Frame(parent)
            bottom_frame.rowconfigure(0, weight=1)
            
            ip = dict_ip(bottom_frame)
            for column, widget in enumerate(ip.values()):
                if column == 0:
                    widget.grid(row=0, column=column, sticky=tk.W, padx=5, pady=5)
                if column == 1:
                    bottom_frame.columnconfigure(column, weight=1)
                    widget.grid(row=0, column=column, sticky=tk.EW, pady=5)
                else:
                    widget.grid(row=0, column=column, sticky=tk.E, padx=5, pady=5)
            
            self.ip = ip
            return bottom_frame

        self.crear_marco()
        
        cuerpo = {
            'top'   : crear_top,
            'middle': crear_middle,
            'bottom': crear_bottom
        }

        self.root.columnconfigure(0, weight=1)
        for index, frame_func in enumerate(cuerpo.values()):
            if index == 1:
                self.root.rowconfigure(index, weight=1)
            frame_func(self.root).grid(row=index, column=0, sticky=tk.NSEW)

    def asignar_eventos(self):
        # Botones
        def button_command(widget:ttk.Widget, press_command, release_command):
            #Click
            widget.bind('<ButtonPress>'  , press_command)
            widget.bind('<ButtonRelease>', release_command)
            #Enter
            widget.bind('<Return>'           , press_command)
            widget.bind('<KeyRelease-Return>', release_command)
            #Barra Espaciadora
            widget.bind('<space>'            , press_command)
            widget.bind('<KeyRelease-space>' , release_command)

        for name, widget in self.control.items():
            widget:ttk.Button
            if name == "up":
                button_command(widget,
                lambda _: (self.robot.avanzar(self.velocidad['movimiento']['value'].get()) if self.robot.is_alive() else ...),
                lambda _: (self.robot.avanzar(0) if self.robot.is_alive() else ...))
            elif name == "left":
                button_command(widget,
                lambda _: (self.robot.izquierda(int(self.velocidad['movimiento']['value'].get()/3)) if self.robot.is_alive() else ...),
                lambda _: (self.robot.izquierda(0) if self.robot.is_alive() else ...))
            elif name == "down":
                button_command(widget,
                lambda _: (self.robot.retroceder(self.velocidad['movimiento']['value'].get()) if self.robot.is_alive() else ...),
                lambda _: (self.robot.retroceder(0) if self.robot.is_alive() else ...))
            elif name == "right":
                button_command(widget,
                lambda _: (self.robot.derecha(int(self.velocidad['movimiento']['value'].get()/3)) if self.robot.is_alive() else ...),
                lambda _: (self.robot.derecha(0) if self.robot.is_alive() else ...))
            elif name == "center":
                button_command(widget,
                lambda _: (self.robot.atacar(self.velocidad['golpe']['r_value'].get()) if self.robot.is_alive() else ...),
                lambda _: (self.robot.perdonar() if self.robot.is_alive() else ...))
        # Mando
        if Linux:
            def conectar_mando():
                try:
                    self.comunicación.put(item=self.mando['combo'].get(), block=False)
                except:
                    print("LAG!")
            combo_refresh = lambda _ : self.mando['combo'].config(values=tuple(device.name for device in get_gamepads()))
            self.mando['combo'].bind("<Button-1>", combo_refresh)
            self.mando['combo'].bind("<Down>", combo_refresh)
            self.mando['button'].config(command=conectar_mando)
        # Mecánica
        def mecánica_click():
            try:
                if self.robot.is_alive():
                    valor = float(self.mecánica['entry'].get())/100 # Centímetros a Metros
                    valor = dis2vel(valor)                          # Metros a Velocidad Porcentual
                    self.robot.atacar(valor)                        # Robot golpea a Velocidad Porcentual
                    self.velocidad['golpe']['w_value'].set(valor)   # Cambia el Indicador de Arriba
                else:
                    print("Robot no Conectado")
            except ValueError:
                print("Entrada inválida")
            except Exception as e:
                print(e)
                
        self.mecánica['entry'].bind('<Return>', lambda _: mecánica_click())
        self.mecánica['button'].bind('<Return>', lambda _: mecánica_click())
        self.mecánica['button'].config(command=mecánica_click)
        # Conexión
        def read_last_ip() -> str:
            output = ""
            with open("IP.tigerbot") as file:
                line = file.readline()
                if "IP =" in line.upper():
                    IP = line.split()[2]
                    for number in IP.split(".", 3):
                        if not (0<=len(number)<=3 and number.isdecimal()):
                            IP = ""
                    output = IP
            return output
        
        def write_last_ip(IP) -> None:
            with open("IP.tigerbot", "w+") as file:
                file.write(f"IP = {IP}")
        
        @hilo
        def crear_robot():
            try:
                IP = self.ip['entry'].get()
                self.robot.conectar(IP)
                print("Conexión exitosa")
                try:
                    write_last_ip(IP)
                except Exception as e:
                    print(e)
            except:
                self.ip['entry'].delete(0, tk.END)
                self.ip['entry'].insert(0, "Error de Conexión")
                print("Error de Conexión")
                
        try:
            self.ip['entry'].insert(0, read_last_ip())
        except FileNotFoundError:
            pass
        self.ip['entry'].bind('<Return>', lambda _: crear_robot())
        self.ip['button'].bind('<Return>', lambda _: crear_robot())
        self.ip['button'].config(command=crear_robot)

    def conectar_mando(self, manilla):
        for dev in (dev for dev in get_gamepads() if manilla == dev.name):
            #TODO: Permitir usar manillas distintas con el mismo nombre.
            if self.manilla_actual is not None and self.manilla_actual.gamepad != dev:
                self.manilla_actual.stop()
                self.manilla_actual = None

            if self.manilla_actual is None:
                self.manilla_actual = (Gamepad(dev))
                print(f"Manilla conectada ({dev.name})")
            else:
                print("Manilla ya seleccionada")

    def desconexión_mando(self):
        print("Manilla Desconectada")
        self.manilla_actual = None
        self.mando['combo'].set("")

    def monitor1(self):
        if not self.comunicación.empty():
            self.conectar_mando(self.comunicación.get())
        if self.manilla_actual is not None and not check(self.manilla_actual):
            self.desconexión_mando()
        if self.robot.is_alive() and self.manilla_actual is not None:
                self.monitor2()
        else:
            ventana.after(200, self.monitor1)
    
    @hilo
    def monitor2(self):
        velocidad_mov_var = self.velocidad['movimiento']['value']
        velocidad_gol_var = self.velocidad['golpe']['r_value']
        self.manilla_actual.start()
        try:
            while self.robot.is_alive():
                if not self.comunicación.empty():
                    self.conectar_mando(self.comunicación.get())
                    self.manilla_actual.start()
                self.manilla_actual.exception()
                if self.manilla_actual._output.poll():
                    key, v = self.manilla_actual._output.recv()
                    vel_mov: int = velocidad_mov_var.get()
                    vel_gol: int = velocidad_gol_var.get()
                    match key:
                        case "cruceta_H":
                            self.robot.izquierda(-v*int(vel_mov/3))
                        case "cruceta_V":
                            self.robot.avanzar(v*vel_mov)
                        case "botón_A":
                            self.robot.retroceder(v*vel_mov)
                        case "botón_B":
                            self.robot.derecha(v*int(vel_mov/3))
                        case "botón_X":
                            self.robot.izquierda(v*int(vel_mov/3))
                        case "botón_Y":
                            self.robot.avanzar(v*vel_mov)
                        case "LB":
                            self.robot.atacar(-v*vel_gol)
                        case "RB":
                            self.robot.atacar(v*vel_gol)
        except OSError as e:
            if str(e) == '[Errno 19] No such device':
                self.desconexión_mando
        except Exception as e:
            print(f"Caught Exception {e}")
        if not self.robot.is_alive():
            self.manilla_actual.stop()
            print("Robot Desconectado")
        self.monitor1()
            
    
if __name__ == "__main__":
    ventana = Tk()
    tigerbot = Interfaz(ventana)
    ventana.mainloop()