#!/usr/bin/env python3
import threading
from functools import wraps
from tkinter import *
from tkinter.ttk import Combobox
from Manilla.Gamepad_Process_Pipe import Gamepad, get_gamepads
from queue import Queue

# Interfaz gráfica para probar la manilla.

def hilo(func):
    @wraps(func)
    def run(*args, **kwargs):
        t = threading.Thread(target=func, args=args, kwargs=kwargs, daemon=True)
        t.start()
        #return t
    return run

ETIQUETAS = (
    ("Botón A: "            ,  "botón_A"),
    ("Botón B: "            ,  "botón_B"),
    ("Botón X: "            ,  "botón_X"),
    ("Botón Y: "            ,  "botón_Y"),
    ("Botón LB: "           ,  "LB"),
    ("Botón RB: "           ,  "RB"),
    ("SELECT: "             ,  "SELECT"),
    ("START: "              ,  "START"),
    ("XBOX: "               ,  "XBOX"),
    ("Botón L: "            ,  "botón_L"),
    ("Botón R: "            ,  "botón_R"),
    ("Palanca Iz. (X): "    ,  "palanca_izquierda_X"),
    ("Palanca Iz. (Y): "    ,  "palanca_izquierda_Y"),
    ("Botón LT: "           ,  "LT"),
    ("Palanca Derecha (X): ",  "palanca_derecha_X"),
    ("Palanca Derecha (Y): ",  "palanca_derecha_Y"),
    ("Botón RT: "           ,  "RT"),
    ("Cruceta Arriba: "     ,  "cruceta_arriba"),
    ("Cruceta Abajo: "      ,  "cruceta_abajo"),
    ("Cruceta Izquierda: "  ,  "cruceta_izquierda"),
    ("Cruceta Derecha: "    ,  "cruceta_derecha")
    )

class Interfaz:
    def __init__(self, ventana_principal:Tk):
        self.ventana         = ventana_principal
        self.frame:Frame     = None
        self.gamepad:Gamepad = None
        self.string_vars     = None
        self.communication   = Queue(5);
        
        ventana_principal.geometry("1080x720")
        ventana_principal.title("Control")

        self.crear_interfaz()
        self.revisar()

    def refresh(self, event=None):
        self.gamepads = get_gamepads()
        self.combo.config(values=tuple(device.name for device in self.gamepads))

    def crear_interfaz(self):
        self.crear_marco()
        self.etiquetas = []
        self.string_vars = {}

        for index, etiqueta in enumerate(ETIQUETAS):
            self.crear_etiqueta(etiqueta[0], index, 0)
            var = StringVar(value="0")
            self.string_vars.update({etiqueta[1]: var})
            self.etiquetas.append(self.crear_etiqueta(var, index, 1))

        self.combo = Combobox(self.frame, state='readonly')
        self.combo.bind("<Button-1>", self.refresh)
        self.combo.bind("<Down>", self.refresh)
        self.combo.grid(row=0, column=2)
        boton = Button(self.frame, text="Seleccionar", command=self.click)
        boton.grid(row=1,column=2)

    def click(self):
        try:
            self.communication.put(item = self.combo.get(), block=False)
        except:
            print("LAG!")
        
    def crear_etiqueta(self, texto, row, column):
        etiqueta = None
        if type(texto) is str:
            etiqueta = Label(self.frame, text=texto)
        elif type(texto) is StringVar:
            etiqueta = Label(self.frame, textvariable=texto)
        etiqueta.grid(row=row, column=column, padx=5, pady=5)
        return etiqueta
    
    def crear_marco(self):
        if self.frame:
            self.frame.destroy()
        self.frame = Frame(ventana_principal)
        self.frame.pack(fill="both", expand=True)

    def click2(self, name):
        for dev in (dev for dev in get_gamepads() if name == dev.name):
            if self.gamepad is not None and self.gamepad.gamepad != dev:
                self.gamepad.stop()
                self.gamepad = None

            if self.gamepad is None:
                self.gamepad = (Gamepad(dev))
                print(dev.name)
            else:
                print("Selected the same gamepad")

    def revisar(self):
        if not self.communication.empty():
            self.click2(self.communication.get())
        
        if self.gamepad:
            self.monitoreo()
        else:
            self.ventana.after(200, self.revisar)
    
    @hilo
    def monitoreo(self):
        self.gamepad.start()
        try:
            while True:
                if not self.communication.empty():
                    self.click2(self.communication.get())
                    self.gamepad.start()
                self.gamepad.exception()
                if self.gamepad._output.poll():
                    (key, value) = self.gamepad._output.recv()
                    value = f"{value:1.4f}" if type(value) is float else value
                    if   key == "cruceta_V":
                        key = "cruceta_arriba" if value == 1 else "cruceta_abajo" if value == -1 else "Y"
                        value = 1
                    elif key == "cruceta_H":
                        key = "cruceta_izquierda" if value == -1 else "cruceta_derecha" if value == 1 else "X"
                        value = 1
                    if key == "Y":
                        self.string_vars['cruceta_arriba'].set(0)
                        self.string_vars['cruceta_abajo'].set(0)
                    elif key == "X":
                        self.string_vars['cruceta_derecha'].set(0)
                        self.string_vars['cruceta_izquierda'].set(0)
                    else:
                        self.string_vars[key].set(value=value)
        except:
            self.gamepad.stop()
            self.gamepad.clear()
            self.gamepad = None
            self.combo.set("")
            self.revisar()

if __name__ == "__main__":
    ventana_principal = Tk()
    interfaz = Interfaz(ventana_principal)
    ventana_principal.mainloop()
