#!/usr/bin/env python3
import evdev
from multiprocessing import Process, Pipe

# Se puede juntar con Gamepad_Process para guardar estados
class Gamepad:
    def __init__(self, gamepad: evdev.InputDevice):
        self.gamepad = gamepad        

        #Lista
        self.claves = (
         'botón_A', 'botón_B' , 'botón_X',
         'botón_Y', 'LB'      , 'RB',
         'SELECT',  'START'   , 'XBOX',
         'botón_L', 'botón_R' , 'palanca_izquierda_X',
         'palanca_izquierda_Y', 'LT',
         'palanca_derecha_X', 'palanca_derecha_Y',
         'RT', 'cruceta_arriba', 'cruceta_abajo',
         'cruceta_izquierda', 'cruceta_derecha')

        self._output, self._input = Pipe(duplex=False)
        #Excepciones
        #self._e                   = None
        self._exception_rcv, self._exception_snd  = Pipe(duplex=False)

        self._monitor_process    = Process(target=self._monitor_controller, daemon=True)

    def exception(self):
        """Levanta una excepción si el proceso de monitoreo cerró inesperadamente"""
        # This version only throws once
        if self._exception_rcv.poll():
            raise self._exception_rcv.recv()
        
        # # This version throws until you start a new process
        # if self._exception_rcv.poll():
        #     self._e = self._exception_rcv.recv()
        #     raise self._e
        # elif self._e is not None:
        #     raise self._e
        # #

    def _monitor_controller(self):
        try:
            match self.gamepad.name:
                case "Microsoft X-Box One S pad":
                    self._monitor_xbox()
                case "Xbox Wireless Controller":
                    self._monitor_xbox_wireless()
                case "Sony Interactive Entertainment Wireless Controller":
                    self._monitor_playstation()
                case "Wireless Controller":
                    self._monitor_playstation()
                case _:
                    raise ValueError(f"Manilla Inapropiada: {self.gamepad.name}")
        except Exception as e:
            self._exception_snd.send(e)
            #El proceso se detiene solo al terminar

    def _monitor_xbox(self):
        #Microsoft X-Box One S pad
        for entrada in self.gamepad.read_loop():
            value  = None
            evento = evdev.categorize(entrada)
            if type(evento) is evdev.KeyEvent:
                evento: evdev.KeyEvent
                value = bool(entrada.value) #bool
                lista = evento.keycode
                if type(lista) is not list:
                    lista = [lista]

                if   "BTN_A"           in lista:
                    self._input.send(('botón_A', value))
                elif "BTN_B"           in lista:
                    self._input.send(('botón_B', value))
                elif "BTN_X"           in lista:
                    self._input.send(('botón_X', value))
                elif "BTN_Y"           in lista:
                    self._input.send(('botón_Y', value))
                elif "BTN_TL"          in lista:
                    self._input.send(('LB', value))
                elif "BTN_TR"          in lista:
                    self._input.send(('RB', value))
                elif "BTN_THUMBL"      in lista:
                    self._input.send(('botón_L', value))
                elif "BTN_THUMBR"      in lista:
                    self._input.send(('botón_R', value))
                elif "BTN_SELECT"      in lista:
                    self._input.send(('SELECT', value))
                elif "BTN_MODE"        in lista:
                    self._input.send(('XBOX', value))
                elif "BTN_START" or "BTN_HOMEPAGE" in lista: #BTN_START en mi Máquina Virtual
                    self._input.send(('START', value))
                else:
                    print("Error: ", lista)
            elif type(evento) is evdev.AbsEvent:
                evento: evdev.AbsEvent
                value = entrada.value
                code  = entrada.code
                match code:
                    case evdev.ecodes.ABS_Z:
                        value = (value-self.gamepad.absinfo(code).min)/(self.gamepad.absinfo(code).max-self.gamepad.absinfo(code).min)
                        self._input.send(('LT', value))
                    case evdev.ecodes.ABS_RZ:
                        value = (value-self.gamepad.absinfo(code).min)/(self.gamepad.absinfo(code).max-self.gamepad.absinfo(code).min)
                        self._input.send(('RT', value))
                    case evdev.ecodes.ABS_X:
                        value = (value-self.gamepad.absinfo(code).min)/(self.gamepad.absinfo(code).max-self.gamepad.absinfo(code).min)*2-1
                        self._input.send(('palanca_izquierda_X', value))
                    case evdev.ecodes.ABS_Y:
                        value = (value-self.gamepad.absinfo(code).min)/(self.gamepad.absinfo(code).max-self.gamepad.absinfo(code).min)*2-1
                        self._input.send(('palanca_izquierda_Y', -value))
                    case evdev.ecodes.ABS_RX:
                        value = (value-self.gamepad.absinfo(code).min)/(self.gamepad.absinfo(code).max-self.gamepad.absinfo(code).min)*2-1
                        self._input.send(('palanca_derecha_X', value))
                    case evdev.ecodes.ABS_RY:
                        value = (value-self.gamepad.absinfo(code).min)/(self.gamepad.absinfo(code).max-self.gamepad.absinfo(code).min)*2-1
                        self._input.send(('palanca_derecha_Y', -value))
                    case evdev.ecodes.ABS_HAT0X:
                        self._input.send(('cruceta_H', value))
                    case evdev.ecodes.ABS_HAT0Y:
                        self._input.send(('cruceta_V', -value))
                    case _:
                        print("Error: ", evento)

    def _monitor_xbox_wireless(self):
        #Xbox Wireless Controller
        for entrada in self.gamepad.read_loop():
            value  = None
            evento = evdev.categorize(entrada)
            if type(evento) is evdev.KeyEvent:
                evento: evdev.KeyEvent
                value = bool(entrada.value)
                lista = evento.keycode
                if type(lista) is not list:
                    lista = [lista]

                if   "BTN_A"           in lista:
                    self._input.send(('botón_A', value))
                elif "BTN_B"           in lista:
                    self._input.send(('botón_B', value))
                elif "BTN_X"           in lista:
                    self._input.send(('botón_X', value))
                elif "BTN_Y"           in lista:
                    self._input.send(('botón_Y', value))
                elif "BTN_TL"          in lista:
                    self._input.send(('LB', value))
                elif "BTN_TR"          in lista:
                    self._input.send(('RB', value))
                elif "BTN_THUMBL"      in lista:
                    self._input.send(('botón_L', value))
                elif "BTN_THUMBR"      in lista:
                    self._input.send(('botón_R', value))
                elif "KEY_BACK"        in lista:
                    self._input.send(('SELECT', value))
                elif "KEY_HOMEPAGE"    in lista:
                    self._input.send(('XBOX', value))
                elif "BTN_START"       in lista:
                    self._input.send(('START', value))
                else:
                    print("Error: ", lista)
            elif type(evento) is evdev.AbsEvent:
                evento: evdev.AbsEvent
                value = entrada.value
                code  = entrada.code
                match code:
                    case evdev.ecodes.ABS_BRAKE:
                        valor = (value-self.gamepad.absinfo(code).min)/(self.gamepad.absinfo(code).max-self.gamepad.absinfo(code).min)
                        self._input.send(('LT', valor))
                    case evdev.ecodes.ABS_GAS:
                        valor = (value-self.gamepad.absinfo(code).min)/(self.gamepad.absinfo(code).max-self.gamepad.absinfo(code).min)
                        self._input.send(('RT', valor))
                    case evdev.ecodes.ABS_X:
                        valor = (value-self.gamepad.absinfo(code).min)/(self.gamepad.absinfo(code).max-self.gamepad.absinfo(code).min)*2-1
                        self._input.send(('palanca_izquierda_X', valor))
                    case evdev.ecodes.ABS_Y:
                        valor = (value-self.gamepad.absinfo(code).min)/(self.gamepad.absinfo(code).max-self.gamepad.absinfo(code).min)*2-1
                        self._input.send(('palanca_izquierda_Y', -valor))
                    case evdev.ecodes.ABS_Z:
                        valor = (value-self.gamepad.absinfo(code).min)/(self.gamepad.absinfo(code).max-self.gamepad.absinfo(code).min)*2-1
                        self._input.send(('palanca_derecha_X', valor))
                    case evdev.ecodes.ABS_RZ:
                        valor = (value-self.gamepad.absinfo(code).min)/(self.gamepad.absinfo(code).max-self.gamepad.absinfo(code).min)*2-1
                        self._input.send(('palanca_derecha_Y', -valor))
                    case evdev.ecodes.ABS_HAT0X:
                        self._input.send(('cruceta_H', value))
                    case evdev.ecodes.ABS_HAT0Y:
                        self._input.send(('cruceta_V', -value))
                    case _:
                        print("Error: ", evento)

    def _monitor_playstation(self):
        # Manilla de PS5 #
        # Sony Interactive Entertainment Wireless Controller
        # Wireless Controller
        MITAD_RANGO = 127.5
        event: evdev.InputEvent   = None
        lista:list                = None
        valor:int                 = None

        for event in self.gamepad.read_loop():
            if event.type == evdev.ecodes.EV_KEY:
                lista = evdev.categorize(event).keycode

                if type(lista) is not list:
                    lista:list = [lista]
                valor = bool(event.value)
                
                if "BTN_NORTH"    in lista:
                    self._input.send(('botón_Y', valor))
                elif "BTN_WEST"   in lista:
                    self._input.send(('botón_X', valor))
                elif "BTN_SOUTH"  in lista:
                    self._input.send(('botón_A', valor))
                elif "BTN_EAST"   in lista:
                    self._input.send(('botón_B', valor))
                elif "BTN_TL"     in lista:
                    self._input.send(('LB', valor))
                elif "BTN_TL2"    in lista:
                    self._input.send(('LT', valor))
                elif "BTN_THUMBL" in lista:
                    self._input.send(('botón_L', valor))
                elif "BTN_TR"     in lista:
                    self._input.send(('RB', valor))
                elif "BTN_TR2"    in lista:
                    self._input.send(('RT', valor))
                elif "BTN_THUMBR" in lista:
                    self._input.send(('botón_R', valor))
                elif "BTN_SELECT" in lista:
                    self._input.send(('SELECT', valor))
                elif "BTN_START"  in lista:
                    self._input.send(('START', valor))
                elif "BTN_MODE"   in lista:
                    self._input.send(('XBOX', valor))
                else:
                    print(lista)
            elif event.type == evdev.ecodes.EV_ABS:
                valor = event.value

                if event.code is evdev.ecodes.ABS_Z:
                    valor = valor/(2*MITAD_RANGO) # 0 - 255
                    self._input.send(('LT', valor))
                elif event.code is evdev.ecodes.ABS_RZ:
                    valor = valor/(2*MITAD_RANGO)
                    self._input.send(('RT', valor))
                elif event.code is evdev.ecodes.ABS_X:
                    valor = (valor/MITAD_RANGO)-1 # 0 - 255 (Izquierda a derecha)
                    self._input.send(('palanca_izquierda_X', valor))
                elif event.code is evdev.ecodes.ABS_Y:
                    valor = (valor/MITAD_RANGO)-1 # 0 - 255 (Arriba a abajo)
                    self._input.send(('palanca_izquierda_Y', -valor))
                elif event.code is evdev.ecodes.ABS_RX:
                    valor = (valor/MITAD_RANGO)-1
                    self._input.send(('palanca_derecha_X', valor))
                elif event.code is evdev.ecodes.ABS_RY:
                    valor = (valor/MITAD_RANGO)-1
                    self._input.send(('palanca_derecha_Y', -valor))
                elif event.code is evdev.ecodes.ABS_HAT0X:
                    self._input.send(('cruceta_H', valor))
                elif event.code is evdev.ecodes.ABS_HAT0Y:
                    self._input.send(('cruceta_V', -valor))

    
    def start(self):
        if not self._monitor_process.is_alive():
            # self._e = None
            try:
                self._monitor_process.start()
            except:
                self._monitor_process = Process(target=self._monitor_controller, daemon=True)
                self._monitor_process.start()
        else:
            print("Advertencia: Proceso ya está corriendo")

    def stop(self):
        if self._monitor_process.is_alive():
            self._monitor_process.terminate()
            self._monitor_process.join()
            # self._monitor_process.close()
        else:
            print("Advertencia: No hay proceso ha detener")

    def clear(self):
        while(self._output.poll()):
            self._output.recv()


def get_gamepad() -> evdev.InputDevice:
    print("Buscando control...")
    device:evdev.InputDevice = evdev.InputDevice(evdev.list_devices()[0])
    print("Encontrado: ", device.name)
    return evdev.InputDevice(device)

def get_gamepads() -> tuple[evdev.InputDevice]:
    return tuple(evdev.InputDevice(device) for device in evdev.list_devices())

def check(mando: Gamepad) -> bool:
    """ Retorna Verdadero si la dirección del dispositivo es válida """
    try:
        evdev.InputDevice(mando.gamepad.path)
        return True
    except:
        return False
        
if __name__ == "__main__":
    import time
    devices = get_gamepads()
    dev = devices[0]
    # for device in devices:
    #     print(f"Device ({device.path} | {device.info}) is a {device.name}")
    print(dev.name)
    x = Gamepad(devices[0])
    # try:
    #     while(True):
    #         x.exception()
    # except:
    #     time.sleep(8)
    #     for device in evdev.list_devices():
    #         print("True" if device == dev.path else "False")
    #     x.start()
    
